--
-- dm_fifo.vhd
-- 
-- This module serves as an interface between
-- the TSB12LV32's DataMover port and video 
-- mode-specific decoding circuitry for one
-- or two Firewire video cameras.
--

library ieee;
library altera_mf;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use altera_mf.altera_mf_components.all;

entity dm_fifo is
generic
(
	CAMERA0_ENABLED		: boolean := true;
	CAMERA1_ENABLED		: boolean := true;
	CAMERA0_CHANNEL		: integer range 0 to 15 := 0;
	CAMERA1_CHANNEL		: integer range 0 to 15 := 1
);
port
(
	-- User clock (any clock, just make sure it's > 25Mhz)
	clock			: in	std_logic;
 	reset_n			: in	std_logic;
 
	-- TSB12LV32 DataMover signals (outside FPGA)
	tm4_fir_dmclk		: in 	std_logic;
	tm4_fir_dmd		: inout std_logic_vector(0 to 15);
	tm4_fir_dmrw		: in 	std_logic;
	tm4_fir_cystart		: in 	std_logic;
	tm4_fir_pktflag		: in 	std_logic;
	tm4_fir_dmpre		: in 	std_logic;
	tm4_fir_dmdone		: in 	std_logic;
	tm4_fir_dmerror		: in 	std_logic;
	tm4_fir_dmready		: out 	std_logic;

	-- Interface to user circuit
	quadlet_data	: out std_logic_vector(31 downto 0); -- 32-bit raw data quadlet
	
	read_ackA	: in std_logic;	-- Assert high to acknowledge current valid quadlet
	quadlet_validA	: out std_logic; -- High if quadlet_data and new_frame are valid
	new_frameA	: out std_logic; -- High if current data quadlet is the first of a new video frame.
	
	read_ackB	: in std_logic; -- Same signals for other camera
	quadlet_validB	: out std_logic; 
	new_frameB	: out std_logic 
);
end dm_fifo;


architecture rtl of dm_fifo is
	-- Temporary copy of tm4_fir_dmd
	signal fir_data		: std_logic_vector(15 downto 0);

	-- Signals for Altera dcfifo megafunction
	signal fifo_in		: std_logic_vector(33 downto 0);
	signal fifo_out		: std_logic_vector(33 downto 0);
	signal fifo_wrreq	: std_logic;
	signal fifo_rdreq	: std_logic;
	signal fifo_rdempty	: std_logic;

	signal which_camera_in	: std_logic;
	signal new_frame_in	: std_logic;
	signal quad_2ndhalf	: boolean;
	signal read_ack		: std_logic;
begin
	-- Instantiate a dual-clock FIFO, connected to
	-- the DataMover 24.576MHz clock on the input
	-- side, and the user clock on the output side.
	-- It is 34 bits wide: 32 bits of data and two
	-- additional control bits.
	fifo : dcfifo
	generic map
	(
		LPM_WIDTH	=> 34,
		LPM_WIDTHU	=> 4,
		LPM_NUMWORDS	=> 16,
		LPM_SHOWAHEAD	=> "ON",
		OVERFLOW_CHECKING => "OFF"
	)
	port map
	(
		data 	=> fifo_in,
		rdclk 	=> clock,
		wrclk 	=> tm4_fir_dmclk,
		wrreq	=> fifo_wrreq,
		rdreq	=> fifo_rdreq,
		aclr	=> not reset_n,
		rdempty => fifo_rdempty,
		q	=> fifo_out
	);


	--
	-- This machine copies data from the DataMover port into the fifo,
	-- synchronized to the DataMover output clock.
	--

	-- Keep DataMover in "read" mode with dmready low
	tm4_fir_dmd <= (others => 'Z');
	fir_data <= tm4_fir_dmd;
	tm4_fir_dmready <= '0';

	process (reset_n, tm4_fir_dmclk)
	begin
		if reset_n = '0' then
			quad_2ndhalf <= false;
			new_frame_in <= '0';
		elsif rising_edge(tm4_fir_dmclk) then
			-- DMDONE is high during the last 16-bits of the packet header, which we snoop
			-- for SYNC and CHANNEL parameters for the upcoming data packet.
			if tm4_fir_dmdone = '1' then
				-- Look at the "channel" field to see which camera this packet came from
				if CAMERA0_ENABLED and CAMERA0_CHANNEL = conv_integer(fir_data(13 downto 8)) then
					which_camera_in <= '0';
				elsif CAMERA1_ENABLED and CAMERA1_CHANNEL = conv_integer(fir_data(13 downto 8)) then
					which_camera_in <= '1';
				end if;

				-- Look at the "sync" field, which is "0001" for the start of a new video frame
				if fir_data(3 downto 0) = "0001" then
					new_frame_in <= '1';
				else
					new_frame_in <= '0';
				end if;

				quad_2ndhalf <= false;
				fifo_wrreq <= '0';
			-- DMRW is high when packet data is being sent through
			elsif tm4_fir_dmrw = '1' then
				-- It takes two 16-bit read cycles to get a 32-bit quadlet
				-- from the DataMover port.
				if not quad_2ndhalf then
					fifo_in(33) <= new_frame_in;
					fifo_in(32) <= which_camera_in;
					fifo_in(31 downto 16) <= fir_data;
					new_frame_in <= '0';
					quad_2ndhalf <= true;
					fifo_wrreq <= '0';
				else
					fifo_in(15 downto 0) <= fir_data;
					quad_2ndhalf <= false;
					fifo_wrreq <= '1';
				end if;
			else
				fifo_wrreq <= '0';
			end if;		
		end if;
	end process;


	--
	-- This reads from the fifo to the output ports,
	-- synchronized to the user clock.
	--
	
	-- Get next quadlet when user asserts read_ack
	read_ack <= (read_ackA and not fifo_out(32)) or
	            (read_ackB and fifo_out(32));
	fifo_rdreq <= read_ack and (not fifo_rdempty);

	-- Generate user output signals from 34-bit fifo word
	quadlet_data <= fifo_out(31 downto 0);
	new_frameA <= fifo_out(33) and (not fifo_out(32)) and not fifo_rdempty;
	new_frameB <= fifo_out(33) and (fifo_out(32)) and not fifo_rdempty;
	quadlet_validA <= (not fifo_rdempty) and (not fifo_out(32));
	quadlet_validB <= (not fifo_rdempty) and (fifo_out(32));
end rtl;
