--
-- firevid_decoder_yuv411.vhd
--
-- This module takes raw 32-bit data quadlets
-- from a Firewire digital camera and generates
-- RGB/row/column signals from the stream.
--
-- This module decodes a 4:1:1 YUV stream
--
-- -------------------------------------------------
--
-- In YUV411, every pixel gets a Y (brightness)
-- value, but U/V components are only updated every
-- 4 pixels. The stream coming from the camera
-- looks like this, with every 3 quadlets giving
-- YUV components for 8 pixels.
--
-- Quadlet 0: U0 Y0 Y1 V0
-- Quadlet 1: Y2 Y3 U4 Y4
-- Quadlet 2: Y5 V4 Y6 Y7
-- ...
-- (similar pattern repeats every 3 quadlets / 8 pixels)
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity firevid_decoder_yuv411 is
generic
(
	WIDTH		: natural := 640;
	HEIGHT		: natural := 480
);
port
(
	clock		: in		std_logic;
 	reset_n		: in		std_logic;

	-- Signals from DataMover-FIFO module
	read_ack	: buffer	std_logic;
	quadlet_valid	: in		std_logic;
	quadlet_data	: in		std_logic_vector(31 downto 0);
	new_frame	: in		std_logic;

	-- User signals
	row		: out		std_logic_vector(9 downto 0);
	column		: out		std_logic_vector(9 downto 0);
	red		: out		std_logic_vector(7 downto 0);
	green		: out		std_logic_vector(7 downto 0);
	blue		: out		std_logic_vector(7 downto 0);
	data_valid	: out		std_logic
);
end firevid_decoder_yuv411;


architecture rtl of firevid_decoder_yuv411 is
	-- Stores the previous data quadlet
	signal last_quadlet : std_logic_vector(31 downto 0);

	signal y_in, u_in, v_in	: std_logic_vector(7 downto 0);
	signal r_out, g_out, b_out : std_logic_vector(7 downto 0);

	type delaychain is array(0 to 3) of std_logic_vector(9 downto 0);
	signal column_in, row_in : std_logic_vector(9 downto 0);
	signal column_r, row_r : delaychain;
	signal data_valid_r : std_logic_vector(3 downto 0);

	signal is_new_frame : boolean;
	signal new_frame_r : std_logic;

	component yuv_to_rgb
	port
	(
		clk : in std_logic;
		y_in, u_in, v_in : in std_logic_vector(7 downto 0);
		r_out, g_out, b_out : out std_logic_vector(7 downto 0)
	);
	end component;
begin
	-- Converts 8-bit YUV to 8-bit RGB
	converter : yuv_to_rgb
	port map
	(
		clock,
		y_in, u_in, v_in,
		r_out, g_out, b_out
	);

	-- Assign the output signals
	red <= r_out;
	green <= g_out;
	blue <= b_out;
	data_valid <= data_valid_r(0);
	column <= column_r(0);
	row <= row_r(0);

	-- This mux extracts 8-bit YUV components from 32-bit
	-- quadlets. The YUV411 format has a pattern that
	-- repeats every 8 pixels / 3 quadlets.
	-- The read_ack signal is also asserted at various times
	-- in order to fetch the next quadlet coming from
	-- the camera.
	process (column_in, quadlet_data, last_quadlet)
	begin
		case column_in(2 downto 0) is
		when "000" =>
			read_ack <= '0';
			y_in <= quadlet_data(23 downto 16);
			u_in <= quadlet_data(31 downto 24);
			v_in <= quadlet_data(7 downto 0);
		when "001" =>
			read_ack <= '1';
			y_in <= quadlet_data(15 downto 8);
			u_in <= quadlet_data(31 downto 24);
			v_in <= quadlet_data(7 downto 0);
		when "010" =>
			read_ack <= '0';
			y_in <= quadlet_data(31 downto 24);
			u_in <= last_quadlet(31 downto 24);
			v_in <= last_quadlet(7 downto 0);
		when "011" =>
			read_ack <= '1';
			y_in <= quadlet_data(23 downto 16);
			u_in <= last_quadlet(31 downto 24);
			v_in <= last_quadlet(7 downto 0);
		when "100" =>
			read_ack <= '0';
			y_in <= last_quadlet(7 downto 0);
			u_in <= last_quadlet(15 downto 8);
			v_in <= quadlet_data(23 downto 16);
		when "101" =>
			read_ack <= '0';
			y_in <= quadlet_data(31 downto 24);
			u_in <= last_quadlet(15 downto 8);
			v_in <= quadlet_data(23 downto 16);
		when "110" =>
			read_ack <= '0';
			y_in <= quadlet_data(15 downto 8);
			u_in <= last_quadlet(15 downto 8);
			v_in <= quadlet_data(23 downto 16);
		when "111" =>
			read_ack <= '1';
			y_in <= quadlet_data(7 downto 0);
			u_in <= last_quadlet(15 downto 8);
			v_in <= quadlet_data(23 downto 16);
		end case;
	end process;
	
	-- Save current value of new_frame so we can edge detect
	is_new_frame <= true when new_frame = '1' and new_frame_r = '0' else false;
	process (reset_n, clock, new_frame)
	begin
		if reset_n = '0' then
			new_frame_r <= '0';
		elsif rising_edge(clock) then
			new_frame_r <= new_frame;
		end if;
	end process;

	-- Delay chain of 4 cycles
	column_r(3) <= (others => '0') when is_new_frame else column_in;
	row_r(3) <= (others => '0') when is_new_frame else row_in;
	data_valid_r(3) <= quadlet_valid;
	process (clock)
	begin
		if rising_edge(clock) then
			column_r(0 to 2) <= column_r(1 to 3);
			row_r(0 to 2) <= row_r(1 to 3);
			data_valid_r(2 downto 0) <= data_valid_r(3 downto 1);
		end if;
	end process;
			
	process (reset_n, clock, new_frame, new_frame_r, quadlet_valid)
	begin
		if reset_n = '0' then
			column_in <= (others => '0');
			row_in <= (others => '0');
		elsif rising_edge(clock) then
			-- When read_ack is 1, we will get a new data
			-- quadlet next cycle, so we need to save the
			-- current one.
			if read_ack = '1' then
				last_quadlet <= quadlet_data;
			end if;

			-- Whenever the incoming data quadlet is
			-- valid, we have pixels to write, and
			-- row/column counters must be advanced.
			if quadlet_valid = '1' then
				-- If this cycle is a new frame, next cycle's
				-- coordinates will be x=1 y=0
				if is_new_frame then
					row_in <= conv_std_logic_vector(0, 10);
					column_in <= conv_std_logic_vector(1, 10);
				elsif column_in = (WIDTH-1) then
					row_in <= row_in + 1;
					column_in <= (others => '0');
				else
					column_in <= column_in + 1;
				end if;
			end if;
		end if;
	end process;
end rtl;
