--
-- firevid_decoder_yuv422.vhd
--
-- This module takes raw 32-bit data quadlets
-- from a Firewire digital camera and generates
-- RGB/row/column signals from the stream.
--
-- This module decodes a 4:2:2 YUV stream
--
-- -------------------------------------------------
--
-- In YUV422, every pixel has a Y (brightness) value,
-- but the U/V color components are only updated
-- every 2 pixels.
--
-- The data stream coming from the camera looks like
-- this:
--
-- Quadlet 0: U0 Y0 V0 Y1
-- Quadlet 1: U2 Y2 V2 Y3
-- ...
-- (every quadlet describes Y/U/V for 2 pixels)
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity firevid_decoder_yuv422 is
generic
(
	WIDTH		: natural := 640;
	HEIGHT		: natural := 480
);
port
(
	clock		: in	std_logic;
 	reset_n		: in	std_logic;

	-- Signals from DataMover-FIFO module
	read_ack	: out	std_logic;
	quadlet_valid	: in	std_logic;
	quadlet_data	: in	std_logic_vector(31 downto 0);
	new_frame	: in	std_logic;

	-- User signals
	row		: buffer	std_logic_vector(9 downto 0);
	column		: buffer	std_logic_vector(9 downto 0);
	red		: out		std_logic_vector(7 downto 0);
	green		: out		std_logic_vector(7 downto 0);
	blue		: out		std_logic_vector(7 downto 0);
	data_valid	: out		std_logic
);
end firevid_decoder_yuv422;


architecture rtl of firevid_decoder_yuv422 is
	signal y_in, u_in, v_in	: std_logic_vector(7 downto 0);
	signal r_out, g_out, b_out : std_logic_vector(7 downto 0);

	signal new_frame_r : std_logic;

	component yuv_to_rgb
	port
	(
		y_in, u_in, v_in : in std_logic_vector(7 downto 0);
		r_out, g_out, b_out : out std_logic_vector(7 downto 0)
	);
	end component;
begin
	-- Converts 8-bit YUV to 8-bit RGB
	converter : yuv_to_rgb
	port map
	(
		y_in, u_in, v_in,
		r_out, g_out, b_out
	);
	
	-- Assign outputs
	red <= r_out;
	green <= g_out;
	blue <= b_out;
	data_valid <= quadlet_valid;

	-- We assert read_ack to acknowledge the current
	-- incoming data quadlet (and thus request a new
	-- data quadlet for next cycle) on every odd pixel.
	read_ack <= column(0);

	-- This multiplexer selects where the YUV values for
	-- the YUVtoRGB converter come from, depending on
	-- whether the current pixel is even or odd-numbered.
	--
	-- For 4:2:2 YUV, every pixel has its own Y, but U/V are
	-- shared between two horizontal neighbor pixels.
	process (column, quadlet_data)
	begin
		if column(0) = '0' then
			y_in <= quadlet_data(23 downto 16);
			u_in <= quadlet_data(31 downto 24);
			v_in <= quadlet_data(15 downto 8);
		else
			y_in <= quadlet_data(7 downto 0);
			u_in <= quadlet_data(31 downto 24);
			v_in <= quadlet_data(15 downto 8);
		end if;
	end process;
	
	-- Save current value of new_frame so we can edge detect
	process (reset_n, clock, new_frame)
	begin
		if reset_n = '0' then
			new_frame_r <= '0';
		elsif falling_edge(clock) then
			new_frame_r <= new_frame;
		end if;
	end process;

	process (reset_n, clock)
	begin
		-- The combination of the latter conditions is there
		-- to ensure that row/column are reset to zero
		-- during only one cycle when new_frame is asserted.
		if reset_n = '0' or (new_frame = '1' and new_frame_r = '0') then
			column <= (others => '0');
			row <= (others => '0');
		elsif rising_edge(clock) then
			-- Update row/column after writing a pixel
			if quadlet_valid = '1' then
				if column = (WIDTH-1) then
					row <= row + 1;
					column <= (others => '0');
				else
					column <= column + 1;
				end if;
			end if;
		end if;
	end process;
end rtl;
